# combo - content management system
# Copyright (C) 2014-2015  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import hashlib
import json
import logging
import requests
import urllib

from django import template
from django.core.cache import cache
from django.db import models
from django.forms import models as model_forms
from django.forms import Select
from django.utils.translation import ugettext_lazy as _

from jsonfield import JSONField

from combo.data.models import CellBase
from combo.data.library import register_cell_class
from combo.utils import NothingInCacheException, sign_url

from .utils import get_wcs_json, is_wcs_enabled, get_wcs_services

logger = logging.getLogger(__name__)

@register_cell_class
class WcsFormCell(CellBase):
    formdef_reference = models.CharField(_('Form'), max_length=150)

    cached_title = models.CharField(_('Title'), max_length=150)
    cached_url = models.URLField(_('URL'))
    cached_json = JSONField(blank=True)

    is_enabled = classmethod(is_wcs_enabled)

    class Meta:
        verbose_name = _('Form Link')

    def get_default_form_class(self):
        from .forms import WcsFormCellForm
        return WcsFormCellForm

    def save(self, *args, **kwargs):
        if self.formdef_reference:
            wcs_key, form_slug = self.formdef_reference.split(':')
            wcs_site = get_wcs_services().get(wcs_key)
            forms_response_json = get_wcs_json(wcs_site, 'api/formdefs/')
            for form in forms_response_json:
                slug = form.get('slug')
                if slug == form_slug:
                    self.cached_title = form.get('title')
                    self.cached_url = form.get('url')
                    self.cached_json = form
        return super(WcsFormCell, self).save(*args, **kwargs)

    def render(self, context):
        formdef_template = template.loader.get_template('combo/wcs/form.html')
        context['slug'] = self.formdef_reference.split(':')[-1]
        context['title'] = self.cached_title
        context['url'] = self.cached_url
        if self.cached_json:
            for attribute in self.cached_json:
                if not attribute in context:
                    context[attribute] = self.cached_json.get(attribute)
        return formdef_template.render(context)

    def get_additional_label(self):
        if not self.cached_title:
            return
        return self.cached_title


class WcsCommonCategoryCell(CellBase):
    is_enabled = classmethod(is_wcs_enabled)
    category_reference = models.CharField(_('Category'), max_length=150)

    cached_title = models.CharField(_('Title'), max_length=150)
    cached_description = models.TextField(_('Description'), blank=True)
    cached_url = models.URLField(_('Cached URL'))

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if self.category_reference:
            wcs_key, category_slug = self.category_reference.split(':')
            wcs_site = get_wcs_services().get(wcs_key)
            categories_response_json = get_wcs_json(wcs_site, 'api/categories/')
            for category in categories_response_json.get('data'):
                slug = category.get('slug')
                if slug == category_slug:
                    self.cached_title = category.get('title')
                    self.cached_description = category.get('description') or ''
                    self.cached_url = category.get('url')
        return super(WcsCommonCategoryCell, self).save(*args, **kwargs)

    def get_additional_label(self):
        if not self.cached_title:
            return
        return self.cached_title


@register_cell_class
class WcsCategoryCell(WcsCommonCategoryCell):

    class Meta:
        verbose_name = _('Category Link')

    def get_default_form_class(self):
        from .forms import WcsCategoryCellForm
        return WcsCategoryCellForm

    def render(self, context):
        if not self.category_reference:
            return ''
        category_template = template.loader.get_template('combo/wcs/category.html')
        context['slug'] = self.category_reference.split(':')[-1]
        context['title'] = self.cached_title
        context['description'] = self.cached_description
        context['url'] = self.cached_url
        return category_template.render(context)


class WcsBlurpMixin(object):
    is_enabled = classmethod(is_wcs_enabled)
    cache_duration = 5

    def get_data(self, context):
        if self.wcs_site:
            try:
                wcs_sites = {self.wcs_site: get_wcs_services()[self.wcs_site]}
            except KeyError:
                # in case of the site disappeared from settings
                return {}
        else:
            wcs_sites = get_wcs_services()

        wcs_sites = wcs_sites.copy()
        for wcs_slug, wcs_site in wcs_sites.items():
            url = wcs_site.get('url')
            if not url.endswith('/'):
                url += '/'
            wcs_site['base_url'] = url

            url += self.api_url
            if not '?' in url:
                url += '?'
            else:
                url += '&'
            url += 'format=json'
            if wcs_site.get('orig') and wcs_site.get('secret'):
                url += '&orig=%s' % wcs_site['orig']
                if context.get('user') and context['user'].is_authenticated():
                    if context.get('request') and hasattr(context['request'], 'session') \
                       and context['request'].session.get('mellon_session'):
                        mellon = context['request'].session['mellon_session']
                        nameid = mellon['name_id_content']
                        url += '&NameID=' + urllib.quote(nameid)
                    elif hasattr(context['user'], 'email') and context['user'].email:
                        url += '&email=' + urllib.quote(context['user'].email)
                else:
                    # add an empty user
                    url += '&NameID='

            cache_key = hashlib.md5(url).hexdigest()
            cache_content = cache.get(cache_key)
            if not cache_content:
                if not context.get('synchronous'):
                    raise NothingInCacheException()
                source_response = requests.get(sign_url(
                        url, wcs_site['secret']),
                        verify=False)
                if source_response.status_code == 200:
                    cache_content = json.loads(source_response.content)
                    cache.set(cache_key, cache_content, self.cache_duration)
                else:
                    logger.error('failed to load %s', url)
                    cache_content = None
            wcs_site['data'] = cache_content

        return wcs_sites

    def get_cell_extra_context(self):
        extra_context = super(WcsBlurpMixin, self).get_cell_extra_context()
        extra_context[self.variable_name] = self.get_data()
        return extra_context


class WcsDataBaseCell(CellBase, WcsBlurpMixin):
    is_enabled = classmethod(is_wcs_enabled)
    wcs_site = models.CharField(_('Site'), max_length=50, blank=True)

    class Meta:
        abstract = True

    def get_default_form_class(self):
        if len(get_wcs_services()) == 1:
            return None
        combo_wcs_sites = [('', _('All'))]
        combo_wcs_sites.extend([(x, y.get('title')) for x, y in get_wcs_services().items()])
        return model_forms.modelform_factory(self.__class__,
                fields=['wcs_site'],
                widgets={'wcs_site': Select(choices=combo_wcs_sites)})

    def render(self, context):
        context[self.variable_name] = self.get_data(context)
        return super(WcsDataBaseCell, self).render(context)


@register_cell_class
class WcsCurrentFormsCell(WcsDataBaseCell):
    api_url = 'api/user/forms'
    variable_name = 'user_forms'

    current_forms = models.BooleanField(_('Current Forms'), default=True)
    done_forms = models.BooleanField(_('Done Forms'), default=False)

    class Meta:
        verbose_name = _('User Forms')

    def get_default_form_class(self):
        form_class = super(WcsCurrentFormsCell, self).get_default_form_class()
        if not form_class:
            return model_forms.modelform_factory(self.__class__,
                    fields=['current_forms', 'done_forms'])
        return form_class

    @property
    def template_name(self):
        if self.current_forms and self.done_forms:
            return 'combo/wcs/user_all_forms.html'
        if self.done_forms:
            return 'combo/wcs/user_done_forms.html'
        return 'combo/wcs/current_forms.html'

    def get_additional_label(self):
        if self.current_forms and self.done_forms:
            return _('All Forms')
        if self.done_forms:
            return _('Done Forms')
        return _('Current Forms')

    def render(self, context):
        context[self.variable_name] = self.get_data(context)
        if not (self.current_forms and self.done_forms):
            for wcs_site in context['user_forms']:
                if not context['user_forms'].get(wcs_site):
                    continue
                if not context['user_forms'][wcs_site].get('data'):
                    continue
                if self.current_forms:
                    context['user_forms'][wcs_site]['data'] = [x for x in
                            context['user_forms'][wcs_site]['data'] if not x.get('form_status_is_endpoint')]
                elif self.done_forms:
                    context['user_forms'][wcs_site]['data'] = [x for x in
                            context['user_forms'][wcs_site]['data'] if x.get('form_status_is_endpoint')]
            context['current_forms'] = context['user_forms'] # legacy
        return super(WcsDataBaseCell, self).render(context)


@register_cell_class
class WcsCurrentDraftsCell(WcsDataBaseCell):
    api_url = 'myspace/drafts'
    variable_name = 'current_drafts'
    template_name = 'combo/wcs/current_drafts.html'

    class Meta:
        verbose_name = _('Current Drafts')


@register_cell_class
class WcsFormsOfCategoryCell(WcsCommonCategoryCell, WcsBlurpMixin):
    ordering = models.CharField(_('Order'), max_length=20,
            default='', blank=True,
            choices=[('', _('Default')),
                     ('alpha', _('Alphabetical')),
                     ('popularity', _('Popularity'))])
    limit = models.PositiveSmallIntegerField(_('Limit'),
            null=True, blank=True)

    class Meta:
        verbose_name = _('Forms of Category')

    variable_name = 'forms'
    template_name = 'combo/wcs/forms_of_category.html'
    cache_duration = 600

    def get_default_form_class(self):
        from .forms import WcsFormsOfCategoryCellForm
        return WcsFormsOfCategoryCellForm

    @property
    def wcs_site(self):
        return self.category_reference.split(':')[0]

    @property
    def api_url(self):
        return self.category_reference.split(':')[1] + '/json'

    def is_relevant(self, user=None):
        return bool(self.category_reference)

    def render(self, context):
        if not self.category_reference:
            return ''
        context[self.variable_name] = self.get_data(context)
        context['slug'] = self.category_reference.split(':')[-1]
        context['title'] = self.cached_title
        context['description'] = self.cached_description
        try:
            context['forms'] = list(context['forms'][self.wcs_site]['data'])
        except (KeyError, TypeError) as e:
            # an error occured in the blurp
            context['forms'] = []
        if self.ordering == 'alpha':
            context['forms'] = sorted(context['forms'], lambda x, y: cmp(x.get('title'), y.get('title')))
        elif self.ordering == 'popularity':
            context['forms'] = sorted(context['forms'], lambda x, y: -cmp(x.get('count'), y.get('count')))

        if self.limit:
            if len(context['forms']) > self.limit:
                context['more_forms'] = True
            context['forms'] = context['forms'][:self.limit]

        return super(WcsFormsOfCategoryCell, self).render(context)


@register_cell_class
class CategoriesCell(WcsDataBaseCell):
    api_url = 'api/categories/?full=on'
    variable_name = 'form_categories'
    template_name = 'combo/wcs/form_categories.html'
    cache_duration = 600

    class Meta:
        verbose_name = _('Form Categories')


@register_cell_class
class TrackingCodeInputCell(CellBase):
    is_enabled = classmethod(is_wcs_enabled)
    wcs_site = models.CharField(_('Site'), max_length=50, blank=True)
    template_name = 'combo/wcs/tracking_code_input.html'

    class Meta:
        verbose_name = _('Tracking Code Input')

    def get_default_form_class(self):
        if len(get_wcs_services()) == 1:
            return None
        combo_wcs_sites = [(x, y.get('title')) for x, y in get_wcs_services().items()]
        return model_forms.modelform_factory(self.__class__,
                fields=['wcs_site'],
                widgets={'wcs_site': Select(choices=combo_wcs_sites)})

    def render(self, context):
        tmpl = template.loader.get_template(self.template_name)
        if not self.wcs_site:
            self.wcs_site = get_wcs_services().keys()[0]
        context['url'] = get_wcs_services().get(self.wcs_site).get('url')
        return tmpl.render(context)
